@@ -16,14 +16,14 @@ import (
 	"github.com/dgraph-io/badger/v4"
 	lru "github.com/hashicorp/golang-lru/v2"
 	"github.com/prometheus/client_golang/prometheus"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/gitaly/config"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/gitaly/storage"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/gitaly/storage/keyvalue"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/gitaly/storage/keyvalue/databasemgr"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/gitaly/storage/mode"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/log"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/safe"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/structerr"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/gitaly/config"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/gitaly/storage"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/gitaly/storage/keyvalue"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/gitaly/storage/keyvalue/databasemgr"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/gitaly/storage/mode"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/log"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/safe"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/structerr"
 )
 
 // ErrPartitionManagerClosed is returned when the PartitionManager stops processing transactions.
@@ -541,7 +541,6 @@ func (sm *StorageManager) startPartition(ctx context.Context, partitionID storag
 				}
 
 				logger := sm.logger.WithField("partition_id", partitionID)
-
 				mgr := sm.partitionFactory.New(
 					ctx,
 					logger,
@@ -666,6 +665,64 @@ func (sm *StorageManager) MaybeAssignToPartition(ctx context.Context, relativePa
 	return sm.partitionAssigner.getPartitionID(ctx, relativePath, "", false)
 }
 
+// MaybeUpdateRepositoryKey ensures that the repositoryKey is added to partition KV
+// only if the repository exists. This method uses a partition transaction to ensure
+// atomicity and conflict detection with concurrent repository operations.
+func (sm *StorageManager) MaybeUpdateRepositoryKey(relativePath string, ptnID storage.PartitionID) (returnErr error) {
+	ctx := context.Background()
+
+	// Get a partition handle to create a transaction
+	ptn, err := sm.GetPartition(ctx, ptnID)
+	if err != nil {
+		return fmt.Errorf("get partition: %w", err)
+	}
+	defer ptn.Close()
+
+	txn, err := ptn.Begin(ctx, storage.BeginOptions{
+		Write:         true,
+		RelativePaths: []string{relativePath},
+	})
+	if err != nil {
+		return fmt.Errorf("begin partition transaction: %w", err)
+	}
+	defer func() {
+		if returnErr != nil {
+			if err := txn.Rollback(ctx); err != nil {
+				returnErr = errors.Join(returnErr, fmt.Errorf("rollback: %w", err))
+			}
+		}
+	}()
+
+	// Verify the repository still exists on disk within the transaction
+	repositoryPath := filepath.Join(sm.path, relativePath)
+	if err := storage.ValidateGitDirectory(repositoryPath); err != nil {
+		if errors.Is(err, fs.ErrNotExist) {
+			// Repository was deleted, don't create the key
+			return nil
+		}
+		return fmt.Errorf("validate git directory: %w", err)
+	}
+
+	kv := txn.KV()
+	key := storage.RepositoryKey(relativePath)
+	_, err = kv.Get(key)
+	if err != nil {
+		if !errors.Is(err, badger.ErrKeyNotFound) {
+			return fmt.Errorf("get KV pair: %w", err)
+		}
+
+		if err := kv.Set(key, nil); err != nil {
+			return fmt.Errorf("set KV pair: %w", err)
+		}
+	}
+
+	if _, err := txn.Commit(ctx); err != nil {
+		return fmt.Errorf("commit partition transaction: %w", err)
+	}
+
+	return nil
+}
+
 // HasPendingWAL checks if a partition has any pending WAL entries by examining the partition's WAL directory.
 // It returns true if there are any WAL entries that needs to be applied, false otherwise.
 func (sm *StorageManager) HasPendingWAL(ctx context.Context, partitionID storage.PartitionID) (bool, error) {
